@@ -39,6 +39,9 @@
 #include <pwd.h>
 #include <time.h>
 
+#include <set>
+#include <list>
+
 #include <libgnomevfs/gnome-vfs-utils.h>
 
 #include "procman.h"
@@ -56,6 +59,19 @@ static guint64 total_time = 1;
 static guint64 total_time_last = 1;
 
 
+
+ProcInfo::List ProcInfo::all;
+
+
+ProcInfo* ProcInfo::find(pid_t pid)
+{
+  Iterator it(ProcInfo::all.find(pid));
+  return (it == ProcInfo::all.end() ? NULL : it->second);
+}
+
+
+
+
 static gint
 sort_ints (GtkTreeModel *model, GtkTreeIter *itera, GtkTreeIter *iterb, gpointer data)
 {
@@ -458,7 +474,7 @@ ProcInfo::~ProcInfo()
   g_free(this->tooltip);
   g_free(this->arguments);
   g_free(this->security_context);
-  g_slist_free(this->children);
+  procman::poison(*this, 0x42);
 }
 
 
@@ -470,27 +486,22 @@ get_process_status (ProcInfo *info, const glibtop_proc_state *buf)
 	{
 	case GLIBTOP_PROCESS_RUNNING:
 		info->status = _("Running");
-		info->is_running = TRUE;
 		break;
 
 	case GLIBTOP_PROCESS_STOPPED:
 		info->status = _("Stopped");
-		info->is_running = FALSE;
 		break;
 
 	case GLIBTOP_PROCESS_ZOMBIE:
 		info->status = _("Zombie");
-		info->is_running = FALSE;
 		break;
 
 	case GLIBTOP_PROCESS_UNINTERRUPTIBLE:
 		info->status = _("Uninterruptible");
-		info->is_running = FALSE;
 		break;
 
 	default:
 		info->status = _("Sleeping");
-		info->is_running = FALSE;
 		break;
 	}
 }
@@ -593,24 +604,6 @@ get_process_memory_info(ProcInfo *info)
 
 
 
-ProcInfo *
-proctable_find_process (guint pid, ProcData *procdata)
-{
-	return static_cast<ProcInfo*>(g_hash_table_lookup(procdata->pids, GINT_TO_POINTER(pid)));
-}
-
-
-static ProcInfo *
-find_parent (ProcData *procdata, guint pid)
-{
-	// init and other kernel processes
-	if (pid == 0)
-		return NULL;
-
-	return proctable_find_process(pid, procdata);
-}
-
-
 static inline unsigned divide(unsigned *q, unsigned *r, unsigned d)
 {
 	*q = *r / d;
@@ -651,8 +644,12 @@ format_duration_for_display(unsigned centiseconds)
 
 
 static void
-update_info_mutable_cols(GtkTreeStore *store, ProcData *procdata, ProcInfo *info)
+update_info_mutable_cols(ProcInfo *info)
 {
+	ProcData * const procdata = ProcData::get_instance();
+	GtkTreeModel *model;
+	model = gtk_tree_view_get_model(GTK_TREE_VIEW(procdata->tree));
+
 	gchar *vmsize, *memres, *memwritable, *memshared, *memxserver,
 		*cpu_time, *start_time, *mem;
 
@@ -670,7 +667,7 @@ update_info_mutable_cols(GtkTreeStore *store, ProcData *procdata, ProcInfo *info
 	   absolute date wouldn't required to be updated on every refresh */
 	start_time = procman_format_date_for_display(info->start_time);
 
-	gtk_tree_store_set (store, &info->node,
+	gtk_tree_store_set(GTK_TREE_STORE(model), &info->node,
 			    COL_STATUS, info->status,
 			    COL_USER, info->user,
 			    COL_VMSIZE, vmsize,
@@ -699,27 +696,26 @@ update_info_mutable_cols(GtkTreeStore *store, ProcData *procdata, ProcInfo *info
 
 
 
-void
+static void
 insert_info_to_tree (ProcInfo *info, ProcData *procdata)
 {
 	GtkTreeModel *model;
 
-	/* Don't show process if it is not running */
-	if ((procdata->config.whose_process == ACTIVE_PROCESSES) &&
-	    (!info->is_running))
-		return;
-
 	model = gtk_tree_view_get_model (GTK_TREE_VIEW (procdata->tree));
-	if (info->parent && procdata->config.show_tree && info->parent->is_visible) {
-		GtkTreePath *parent_node = gtk_tree_model_get_path (model, &info->parent->node);
 
-		gtk_tree_store_insert (GTK_TREE_STORE (model), &info->node, &info->parent->node, 0);
-		if (!gtk_tree_view_row_expanded (GTK_TREE_VIEW (procdata->tree), parent_node))
-			gtk_tree_view_expand_row (GTK_TREE_VIEW (procdata->tree),
-						  parent_node,
-						  FALSE);
+	if (procdata->config.show_tree) {
 
-		gtk_tree_path_free (parent_node);
+	  ProcInfo *parent = ProcInfo::find(info->ppid);
+
+	  if (parent) {
+	    GtkTreePath *parent_node = gtk_tree_model_get_path(model, &parent->node);
+	    gtk_tree_store_insert(GTK_TREE_STORE(model), &info->node, &parent->node, 0);
+
+	    if (!gtk_tree_view_row_expanded(GTK_TREE_VIEW(procdata->tree), parent_node))
+	      gtk_tree_view_expand_row(GTK_TREE_VIEW(procdata->tree), parent_node, FALSE);
+	    gtk_tree_path_free(parent_node);
+	  } else
+	    gtk_tree_store_insert(GTK_TREE_STORE(model), &info->node, NULL, 0);
 	}
 	else
 		gtk_tree_store_insert (GTK_TREE_STORE (model), &info->node, NULL, 0);
@@ -738,266 +734,196 @@ insert_info_to_tree (ProcInfo *info, ProcData *procdata)
 			    COL_PID, info->pid,
 			    COL_SECURITYCONTEXT, info->security_context,
 			    -1);
-
-	update_info_mutable_cols(GTK_TREE_STORE (model), procdata, info);
-
-	info->is_visible = TRUE;
 }
 
 
 /* Removing a node with children - make sure the children are queued
 ** to be readded.
 */
+template<typename List>
 static void
-remove_children_from_tree (ProcData *procdata, GtkTreeModel *model,
-			   GtkTreeIter *parent, GPtrArray *addition_list)
+remove_info_from_tree (ProcData *procdata, GtkTreeModel *model,
+		       ProcInfo *current, List &orphans, unsigned lvl = 0)
 {
-	do {
-		ProcInfo *child_info;
-		GtkTreeIter child;
+  GtkTreeIter child_node;
 
-		if (gtk_tree_model_iter_children (model, &child, parent))
-			remove_children_from_tree (procdata, model, &child, addition_list);
+  if (std::find(orphans.begin(), orphans.end(), current) != orphans.end()) {
+    procman_debug("[%u] %d already removed from tree", lvl, int(current->pid));
+    return;
+  }
 
-		gtk_tree_model_get (model, parent, COL_POINTER, &child_info, -1);
-		if (child_info) {
-			if (procdata->selected_process == child_info)
-				procdata->selected_process = NULL;
-			g_ptr_array_add(addition_list, child_info);
-			child_info->is_visible = FALSE;
-		}
-	} while (gtk_tree_model_iter_next (model, parent));
+  procman_debug("[%u] pid %d, %d children", lvl, int(current->pid),
+		gtk_tree_model_iter_n_children(model, &current->node));
+
+  // it is not possible to iterate&erase over a treeview so instead we
+  // just pop one child after another and recursively remove it and
+  // its children
+
+  while (gtk_tree_model_iter_children(model, &child_node, &current->node)) {
+    ProcInfo *child = 0;
+    gtk_tree_model_get(model, &child_node, COL_POINTER, &child, -1);
+    remove_info_from_tree(procdata, model, child, orphans, lvl + 1);
+  }
+
+  g_assert(not gtk_tree_model_iter_has_child(model, &current->node));
+
+  if (procdata->selected_process == current)
+    procdata->selected_process = NULL;
+
+  orphans.push_back(current);
+  gtk_tree_store_remove(GTK_TREE_STORE(model), &current->node);
+  procman::poison(current->node, 0x69);
 }
 
 
-void
-remove_info_from_tree (ProcInfo *info, ProcData *procdata)
-{
-	GtkTreeModel *model;
-
-	g_assert(info);
-
-	if (!info->is_visible)
-		return;
-
-	model = gtk_tree_view_get_model (GTK_TREE_VIEW (procdata->tree));
-
-	if (procdata->selected_process == info)
-		procdata->selected_process = NULL;
-
-	gtk_tree_store_remove (GTK_TREE_STORE (model), &info->node);
-
-	info->is_visible = FALSE;
-}
-
-
-static void
-remove_info_from_list (ProcInfo *info, ProcData *procdata)
-{
-	GSList *child;
-	ProcInfo * const parent = info->parent;
-
-	/* Remove info from parent list */
-	if (parent)
-		parent->children = g_slist_remove (parent->children, info);
-
-	/* Add any children to parent list */
-	for(child = info->children; child; child = g_slist_next (child)) {
-		ProcInfo *child_info = static_cast<ProcInfo*>(child->data);
-		child_info->parent = parent;
-	}
-
-	if(parent) {
-		parent->children = g_slist_concat(parent->children,
-						  info->children);
-		info->children = NULL;
-	}
-
-	/* Remove from list */
-	procdata->info = g_list_remove (procdata->info, info);
-	g_hash_table_remove(procdata->pids, GINT_TO_POINTER(info->pid));
-	delete info;
-}
-
 
 static void
 update_info (ProcData *procdata, ProcInfo *info)
 {
 	glibtop_proc_state procstate;
+	glibtop_proc_uid procuid;
+	glibtop_proc_time proctime;
 
 	glibtop_get_proc_state (&procstate, info->pid);
+	glibtop_get_proc_uid (&procuid, info->pid);
+	glibtop_get_proc_time (&proctime, info->pid);
+
 	get_process_status (info, &procstate);
 
-	if (procdata->config.whose_process == ACTIVE_PROCESSES)	{
+	get_process_memory_info(info);
 
-		/* process started running */
-		if (info->is_running && (!info->is_visible)) {
-			insert_info_to_tree (info, procdata);
-			return;
-		}
-		/* process was running but not anymore */
-		else if ((!info->is_running) && info->is_visible) {
-			remove_info_from_tree (info, procdata);
-			return;
-		}
-		else if (!info->is_running)
-			return;
-	}
+	get_process_user(procdata, info, procstate.uid);
 
-	if (info->is_visible) {
-		GtkTreeModel *model;
-		glibtop_proc_uid procuid;
-		glibtop_proc_time proctime;
+	info->pcpu = (proctime.rtime - info->cpu_time_last) * 100 / total_time;
+	info->pcpu = MIN(info->pcpu, 100);
 
-		glibtop_get_proc_uid (&procuid, info->pid);
-		glibtop_get_proc_time (&proctime, info->pid);
-
-		get_process_memory_info(info);
-		get_process_user(procdata, info, procstate.uid);
-
-		info->pcpu = (proctime.rtime - info->cpu_time_last) * 100 / total_time;
-		info->pcpu = MIN(info->pcpu, 100);
-
-		info->cpu_time_last = proctime.rtime;
-		info->nice = procuid.nice;
-
-
-		model = gtk_tree_view_get_model (GTK_TREE_VIEW (procdata->tree));
-
-		update_info_mutable_cols(GTK_TREE_STORE (model), procdata, info);
-	}
+	info->cpu_time_last = proctime.rtime;
+	info->nice = procuid.nice;
+	info->ppid = procuid.ppid;
 }
 
 
 ProcInfo::ProcInfo(pid_t pid)
-  : path(NULL),
-    parent(NULL),
-    children(NULL),
-    pixbuf(NULL),
+  : pixbuf(NULL),
     tooltip(NULL),
     name(NULL),
     user(NULL),
     arguments(NULL),
     status(NULL),
-    security_context(NULL)
+    security_context(NULL),
+    pid(pid),
+    uid(-1)
 {
 	ProcInfo * const info = this;
 	glibtop_proc_state procstate;
 	glibtop_proc_time proctime;
-	glibtop_proc_uid procuid;
 	glibtop_proc_args procargs;
 	gchar** arguments;
 
-	info->pid = pid;
-	info->uid = -1;
-
 	glibtop_get_proc_state (&procstate, pid);
-	glibtop_get_proc_uid (&procuid, pid);
 	glibtop_get_proc_time (&proctime, pid);
 	arguments = glibtop_get_proc_argv (&procargs, pid, 0);
 
 	/* FIXME : wrong. name and arguments may change with exec* */
 	get_process_name (info, procstate.cmd, arguments[0]);
-	get_process_user(ProcData::get_instance(), info, procstate.uid);
 
 	info->tooltip = g_strjoinv(" ", arguments);
 	info->arguments = g_strescape(info->tooltip, "\\\"");
 	g_strfreev(arguments);
 
-	info->pcpu = 0;
 	info->cpu_time_last = proctime.rtime;
 	info->start_time = proctime.start_time;
-	info->nice = procuid.nice;
 
-	get_process_memory_info(info);
-	get_process_status (info, &procstate);
 	get_process_selinux_context (info);
 
 	info->pixbuf = ProcData::get_instance()->pretty_table.get_icon(info->name, pid);
-
-	info->parent = find_parent(ProcData::get_instance(), procuid.ppid);
-	if (info->parent) {
-		info->parent->children = g_slist_prepend (info->parent->children, info);
-	}
-
-	info->is_visible = FALSE;
 }
 
 
 
-static void cb_exclude(ProcInfo* info, GPtrArray *addition)
-{
-	g_ptr_array_remove_fast (addition, info);
-}
-
 
 static void
 refresh_list (ProcData *procdata, const unsigned *pid_list, const guint n)
 {
-	GHashTable *pid_hash;
-	GList *list;
-	GPtrArray *addition_list = g_ptr_array_new ();
-	GPtrArray *removal_list = g_ptr_array_new ();
+  typedef std::list<ProcInfo*> ProcList;
+  ProcList addition;
+
 	GtkTreeModel *model = gtk_tree_view_get_model (GTK_TREE_VIEW (procdata->tree));
 	guint i;
 
-	/* Add or update processes */
+	// Add or update processes in the process list
 	for(i = 0; i < n; ++i) {
-		ProcInfo *info;
+		ProcInfo *info = ProcInfo::find(pid_list[i]);
 
-		info = proctable_find_process (pid_list[i], procdata);
 		if (!info) {
 			info = new ProcInfo(pid_list[i]);
-			g_ptr_array_add(addition_list, info);
-			procdata->info = g_list_prepend (procdata->info, info);
-			g_hash_table_insert(procdata->pids, GINT_TO_POINTER(info->pid), info);
-		}
-		else {
-			update_info (procdata, info);
+			ProcInfo::all[info->pid] = info;
+			addition.push_back(info);
 		}
+
+		update_info (procdata, info);
 	}
 
 
-	/* Remove processes */
+	// Remove dead processes from the process list and from the
+	// tree. children are queued to be readded at the right place
+	// in the tree.
 
-	/* use a hash for fast lookup
-	   !NULL as data for every key */
+	const std::set<pid_t> pids(pid_list, pid_list + n);
 
-	pid_hash = g_hash_table_new(NULL, NULL);
-	for(i = 0; i < n; ++i)
-		g_hash_table_insert(pid_hash,
-				    GINT_TO_POINTER(pid_list[i]),
-				    GINT_TO_POINTER(1 /* !NULL */));
+	ProcInfo::Iterator it(ProcInfo::begin());
 
+	while (it != ProcInfo::end()) {
+	  ProcInfo * const info = it->second;
+	  ProcInfo::Iterator next(it);
+	  ++next;
 
-	for(list = procdata->info; list; list = g_list_next (list)) {
-		ProcInfo * const info = static_cast<ProcInfo*>(list->data);
+	  if (pids.find(info->pid) == pids.end()) {
+	    procman_debug("ripping %d", info->pid);
+	    remove_info_from_tree(procdata, model, info, addition);
+	    addition.remove(info);
+	    ProcInfo::all.erase(it);
+	    delete info;
+	  }
 
-		if(!g_hash_table_lookup(pid_hash, GINT_TO_POINTER(info->pid)))
-		{
-			g_ptr_array_add (removal_list, info);
-			/* remove all children from tree */
-			if (info->is_visible) {
-				GtkTreeIter child;
-				if (gtk_tree_model_iter_children (model, &child, &info->node))
-					remove_children_from_tree (procdata, model, &child, addition_list);
-			}
-		}
+	  it = next;
 	}
 
-	g_hash_table_destroy(pid_hash);
+	// insert process in the tree. walk through the addition list
+	// (new process + process that have a new parent). This loop
+	// handles the dependencies because we cannot insert a process
+	// until its parent is in the tree.
 
-	g_ptr_array_foreach(removal_list, (GFunc) cb_exclude, addition_list);
+	std::set<pid_t> in_tree(pids);
 
-	/* Add or remove processes from the tree */
-	g_ptr_array_foreach(removal_list,  (GFunc) remove_info_from_tree, procdata);
-	g_ptr_array_foreach(addition_list, (GFunc) insert_info_to_tree  , procdata);
+	for (ProcList::iterator it(addition.begin()); it != addition.end(); ++it)
+	  in_tree.erase((*it)->pid);
 
-	/* Remove processes from the internal GList */
-	g_ptr_array_foreach(removal_list,  (GFunc) remove_info_from_list, procdata);
+	unsigned rounds = 0;
 
-	g_ptr_array_free (addition_list, TRUE);
-	g_ptr_array_free (removal_list, TRUE);
+	while (not addition.empty()) {
+	  procman_debug("looking for %d parents", int(addition.size()));
+	  ProcList::iterator it(addition.begin());
+
+	  while (it != addition.end()) {
+	    procman_debug("looking for %d's parent", int((*it)->pid));
+	    if ((*it)->ppid == 0 or in_tree.find((*it)->ppid) != in_tree.end()) {
+	      insert_info_to_tree(*it, procdata);
+	      in_tree.insert((*it)->pid);
+	      it = addition.erase(it);
+	      rounds = 0;
+	    } else
+	      ++it;
+	  }
+
+	  rounds++;
+
+	  // dead loop
+	  g_assert(rounds <= 2);
+	}
+
+	for (ProcInfo::Iterator it(ProcInfo::begin()); it != ProcInfo::end(); ++it)
+		update_info_mutable_cols(it->second);
 }
 
 
@@ -1011,10 +937,15 @@ proctable_update_list (ProcData * const procdata)
 
 	switch (procdata->config.whose_process) {
 	case ALL_PROCESSES:
-	case ACTIVE_PROCESSES:
 		which = GLIBTOP_KERN_PROC_ALL;
 		arg = 0;
 		break;
+
+	case ACTIVE_PROCESSES:
+		which = GLIBTOP_KERN_PROC_ALL | GLIBTOP_EXCLUDE_IDLE;
+		arg = 0;
+		break;
+
 	default:
 		which = GLIBTOP_KERN_PROC_UID;
 		arg = getuid ();
@@ -1068,20 +999,10 @@ proctable_clear_tree (ProcData * const procdata)
 void
 proctable_free_table (ProcData * const procdata)
 {
-	GList *list = procdata->info;
+  for (ProcInfo::Iterator it(ProcInfo::begin()); it != ProcInfo::end(); ++it)
+    delete it->second;
 
-	while (list)
-	{
-		ProcInfo *info = static_cast<ProcInfo*>(list->data);
-		delete info;
-		list = g_list_next (list);
-	}
-
-	g_list_free (procdata->info);
-	procdata->info = NULL;
-
-	g_hash_table_destroy(procdata->pids);
-	procdata->pids = g_hash_table_new(NULL, NULL);
+  ProcInfo::all.clear();
 }
 
 
@@ -1112,11 +1033,9 @@ ProcInfo::set_icon(GdkPixbuf* icon)
   this->pixbuf = icon;
   g_object_ref(icon);
 
-  if (this->is_visible) {
-    GtkTreeModel *model;
-    model = gtk_tree_view_get_model(GTK_TREE_VIEW(ProcData::get_instance()->tree));
-    gtk_tree_store_set(GTK_TREE_STORE(model), &this->node,
-		       COL_PIXBUF, this->pixbuf,
-		       -1);
-  }
+  GtkTreeModel *model;
+  model = gtk_tree_view_get_model(GTK_TREE_VIEW(ProcData::get_instance()->tree));
+  gtk_tree_store_set(GTK_TREE_STORE(model), &this->node,
+		     COL_PIXBUF, this->pixbuf,
+		     -1);
 }
