@@ -76,21 +76,21 @@ SI_gnome_vfs_format_file_size_for_display (GnomeVFSFileSize size)
 	if (size < (GnomeVFSFileSize) KIBIBYTE_FACTOR) {
 		return g_strdup_printf (dngettext(GETTEXT_PACKAGE, "%u byte", "%u bytes",(guint) size), (guint) size);
 	} else {
-		gdouble displayed_size;
+		guint factor;
+		const char* format;
 
 		if (size < (GnomeVFSFileSize) MEBIBYTE_FACTOR) {
-			displayed_size = (gdouble) size / KIBIBYTE_FACTOR;
-			return g_strdup_printf (_("%.1f KiB"),
-						       displayed_size);
+		  factor = KIBIBYTE_FACTOR;
+		  format = N_("%.1f KiB");
 		} else if (size < (GnomeVFSFileSize) GIBIBYTE_FACTOR) {
-			displayed_size = (gdouble) size / MEBIBYTE_FACTOR;
-			return g_strdup_printf (_("%.1f MiB"),
-						       displayed_size);
+		  factor = MEBIBYTE_FACTOR;
+		  format = N_("%.1f MiB");
 		} else {
-			displayed_size = (gdouble) size / GIBIBYTE_FACTOR;
-			return g_strdup_printf (_("%.1f GiB"),
-						       displayed_size);
+		  factor = GIBIBYTE_FACTOR;
+		  format = N_("%.1f GiB");
 		}
+
+		return g_strdup_printf(_(format), size / (double)factor);
 	}
 }
 
@@ -177,7 +177,8 @@ static double get_relative_time(void)
 
 
 void
-procman_debug(const char *format, ...)
+procman_debug_real(const char *file, int line, const char *func,
+		   const char *format, ...)
 {
 	va_list args;
 	char *msg;
@@ -189,8 +190,24 @@ procman_debug(const char *format, ...)
 	msg = g_strdup_vprintf(format, args);
 	va_end(args);
 
-	g_debug("[%0.6f] %s", get_relative_time(), msg);
+	g_debug("[%.3f %s:%d %s] %s", get_relative_time(), file, line, func, msg);
 
 	g_free(msg);
 }
 
+
+
+namespace procman
+{
+  void size_cell_data_func(GtkTreeViewColumn *col, GtkCellRenderer *renderer,
+			   GtkTreeModel *model, GtkTreeIter *iter,
+			   gpointer user_data)
+  {
+    guint64 size;
+    gtk_tree_model_get(model, iter, GPOINTER_TO_UINT(user_data), &size, -1);
+
+    char *str = SI_gnome_vfs_format_file_size_for_display(size);
+    g_object_set(renderer, "text", str, NULL);
+    g_free(str);
+  }
+}
